/* #includes */ /*{{{C}}}*//*{{{*/
#ifndef NO_POSIX_SOURCE
#undef _POSIX_SOURCE
#define _POSIX_SOURCE   1
#undef _POSIX_C_SOURCE
#define _POSIX_C_SOURCE 2
#endif

#ifdef DMALLOC
#include "dmalloc.h"
#endif

#include <assert.h>
#include <ctype.h>
extern char *strdup(const char* s);
extern double strtod(const char *nptr, char **endptr); /* SunOS 4 hack */
extern long double strtold(const char *nptr, char **endptr); /* SunOS 4 hack */
#include <string.h>


#include "default.h"
#include "func.h"
#include "main.h"
#include "misc.h"
#include "scanner.h"
/*}}}*/

const char *Type_Name[] =
  { [EMPTY] = "EMPTY", [STRING] = "STRING", [FLOAT] = "FLOAT", [INT] = "INT",
    [OPERATOR] = "OPERATOR", [LIDENT] = "LABEL", [FIDENT] = "FUNCTION",
    [LOCATION] = "LOCATION", [FUNCALL] = "FUNCTION-CALL", [EEK] = "ERROR"
  };

const char *Op_Name[] =
  { [PLUS] = "+", [MINUS] = "-", [MUL] = "*", [DIV] = "/",
    [OP] = "(", [CP] = ")", [COMMA] = ",",
    [LT] = "<", [LE] = "<=", [GE] = ">=", [GT] = ">",
    [ISEQUAL] = "==", [ABOUTEQ] = "~=", [NE] = "!=",
    [POW] = "^", [MOD] = "%"
  };

/* loc_in_box  -- returns true if test is in the box determined by b and c */
bool loc_in_box(const Location test,
		       const Location b, const Location c)
{
  for (Dimensions dim = X; dim < HYPER; ++dim)
  {
    if (test[dim] < b[dim] && test[dim] < c[dim]) return false;
    if (test[dim] > b[dim] && test[dim] > c[dim]) return false;
  }
  return true;
}

/* cleartoken - Initialize all of the memory of a token */ /*{{{*/
void cleartoken(Token* tok)
{
  tok->type = EMPTY;
  tok->u.location[0] = 0;
  tok->u.location[1] = 0;
  tok->u.location[2] = 0;
}

/* duperror - Sets tok to an error and strdups the message into place */
Token duperror(Token* tok, const char* erro)
{
  tok->type = EEK;
  tok->u.err = strdup(erro);
  return *tok;
}

/* charstring -- match quoted string and return token */ /*{{{*/
static Token *charstring(const char **s)
{
  const char *r;

  r=*s;
  if (**s=='"')
  {
    ++(*s);
    while (**s!='\0' && **s!='"') if (**s=='\\' && *((*s)+1)!='\0') (*s)+=2; else ++(*s);
    if (**s=='\0') { *s=r; return 0; }
    else 
    {
      Token *n;
      char *t;

      ++(*s);
      n=malloc(sizeof(Token));
      n->type=STRING;
      t=n->u.string=malloc((size_t)(*s-r));
      /* Clean string of quotes.  This may waste a few bytes, so? */
      ++r;
      while (r<(*s-1)) if (*r=='\\') { *t++=*(r+1); r+=2; } else *t++=*r++;
      *t='\0';
      return n;
    }
  }
  else return (Token*)0;
}
/*}}}*/

/* integer    -- match an unsigned integer and return token */ /*{{{*/
static Token *integer(const char **s)
{
  const char *r;
  long i;

  r=*s;
  i=posnumber(r,s);
  if (*s!=r && **s!='.' && **s!='e')
  {
    Token *n;

    n=malloc(sizeof(Token));
    n->type=INT;
    n->u.integer=i;
    return n;
  }
  else { *s=r; return (Token*)0; }
}
/*}}}*/

/* flt        -- match a floating point number */ /*{{{*/
static Token *flt(const char **s)
{
  /* variables */ /*{{{*/
  const char *t;
  char *end;
  Token *n;
  FltT x;
  /*}}}*/

  t=*s;
  x = STRTOFLT(t,&end);
  *s = end;
  if (t!=*s && dblfinite(x)==(const char*)0)
  {
    n=malloc(sizeof(Token));
    n->type=FLOAT;
    n->u.flt=x;
    return n;
  }
  else
  {
    *s=t;
    return (Token*)0;
  }
}
/*}}}*/

/* op   -- match an op and return token */ /*{{{*/
static Token *op(const char **s)
{
  Token *n;
  Operator op;

  switch (**s)
  {
    case '+': op=PLUS; break;
    case '-': op=MINUS; break;
    case '*': op=MUL; break;
    case '/': op=DIV; break;
    case '%': op=MOD; break;
    case '(': op=OP; break;
    case ')': op=CP; break;
    case ',': op=COMMA; break;
    case '^': op=POW; break;
    case '<': if (*(*s+1)=='=') { ++(*s); op=LE; } else op=LT; break;
    case '>': if (*(*s+1)=='=') { ++(*s); op=GE; } else op=GT; break;
    case '=': if (*(*s+1)=='=') { ++(*s); op=ISEQUAL; } else return (Token*)0; break;
    case '~': if (*(*s+1)=='=') { ++(*s); op=ABOUTEQ; } else return (Token*)0; break;
    case '!': if (*(*s+1)=='=') { ++(*s); op=NE; } else return (Token*)0; break;
    default: return (Token*)0;
  }
  n=malloc(sizeof(Token));
  n->type=OPERATOR;
  n->u.op=op;
  ++(*s);
  return n;
}
/*}}}*/
/* ident      -- match an identifier and return token */ /*{{{*/
static Token *ident(const char **s)
{
  const char *begin;
  Token *result;
  
  if (isalpha((int)**s) || **s=='_' || **s=='@' || **s=='&' || **s=='.' || **s=='$')
  {
    int fident;

    begin=*s; ++(*s);
    while (isalpha((int)**s) || **s=='_' || **s=='@' || **s=='&' || **s=='.' || **s=='$' || isdigit((int)**s)) ++(*s);
    result=malloc(sizeof(Token));
    if ((fident = identcode(begin,(size_t)(*s-begin))) == NOT_A_FUNCTION)
    {
      result->type=LIDENT;
      result->u.lident=malloc((size_t)(*s-begin+1));
      (void)strncpy(result->u.lident,begin,(size_t)(*s-begin));
      result->u.lident[*s-begin]='\0';
    }
    else
    {
      result->type=FIDENT;
      result->u.fident=fident;
    }
    return result;
  }
  return (Token*)0;
}
/*}}}*/

/* scan       -- scan string into tokens */ /*{{{*/
Token **scan(const char **s)
{
  /* variables */ /*{{{*/
  Token **na,*n;
  const char *r;
  /*}}}*/

  /* compute number of tokens */ /*{{{*/
  r = *s;
  while (*r == ' ') ++r;
  int i = 0;
  for (; *r != '\0'; ++i)
  {
    const char *or;

    or = r;
    while (*r == ' ') ++r;
    n = charstring(&r);
    if (n == NULLTOKEN) n = op(&r);
    if (n == NULLTOKEN) n = integer(&r);
    if (n == NULLTOKEN) n = flt(&r);
    if (n == NULLTOKEN) n = ident(&r);
    if (or == r) { *s = r; return EMPTY_TVEC; }
  }
  /*}}}*/
  /* allocate token space */ /*{{{*/
  na = malloc(sizeof(Token*)*(i+1));
  /*}}}*/
  /* store tokens */ /*{{{*/
  r = *s;
  while (*r==' ') ++r;
  for (int j = 0; j < i; ++j)
  {
    while (*r == ' ') ++r;
    n = charstring(&r);
    if (n == NULLTOKEN) n = op(&r);
    if (n == NULLTOKEN) n = integer(&r);
    if (n == NULLTOKEN) n = flt(&r);
    if (n == NULLTOKEN) n = ident(&r);
    na[j] = n;
  }
  na[i] = NULLTOKEN;
  /*}}}*/
  return na;
}
/*}}}*/
